;; -*- Mode: Emacs-Lisp -*-
;; PRCS - The Project Revision Control System
;; Copyright (C) 1997  Josh MacDonald
;;
;; Most of this file by Jesse Glick <jglick@sig.bsh.com>; concepts
;; also due to Zack Weinberg <zack@rabi.phys.columbia.edu>, Dan
;; Nicolaescu <done@ece.arizona.edu>; XEmacs code from Vincent Hascoet
;; <hascoet@lep-philips.fr>, though others have helped too. Current
;; maintainer Jesse Glick <jglick@sig.bsh.com>; please contact
;; directly or CC me. Patches are welcome, but please make sure the
;; PrcsModeVersion is indicated.
;;
;; This program is free software; you can redistribute it and/or
;; modify it under the terms of the GNU General Public License
;; as published by the Free Software Foundation; either version 2
;; of the License, or (at your option) any later version.
;;
;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program; if not, write to the Free Software
;; Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

;; PRCS utility functionality: mainly, PRCS Major Mode (for project
;; files)

;; Reasonable .emacs fodder:
;;
;; (require 'prcs) ;; simple solution
;;
;; This is less useful because you would have to visit the project
;; file first, it will not notify you that a file is controlled:
;;
;; (require 'cl)
;; (pushnew '("\\.prj\\'" . prcs-mode) auto-mode-alist)
;; (autoload 'prcs-mode "prcs" "PRCS project-file mode." t)

;; $PrcsModeVersion: 1.44 on Sat, 28 Nov 1998 16:43:31 -0500 $
;;
;; TODO:
;;
;; Asynch process killed by timeout should send SIGTERM & wait a
;; moment before killing for real, to reduce chance of spurious locks
;; being left around repository.
;;
;; Skip PRCS-controlled-mode if over a PRCS-controlled file, but PRCS
;; is not installed. On a related note, when visiting a p-file whose
;; status check fails, or if there is some problem (should maybe
;; attempt a prcs info -r. at once) that indicates this project might
;; not actually be in local repo (e.g. if you were taking a look at
;; the PRCS sources), should just disable PRCS-controlled mode on
;; p-file and other project files.
;;
;; Possibly have a global var to turn off all asynch process runs (can
;; just modify `prcs-command') in case local Emacs does not handle
;; them well enough. Have had difficult-to-diagnose problems with
;; asynch processes on occasion.
;;
;; Confirm XEmacs compatibility.
;;
;; A number of things could be stored as buffer-local variables that
;; currently are not: project file path; project relative path; maybe
;; others. These would not have serious consistency problems since you
;; do not expect them to change within a buffer's lifetime.
;;
;; Handling of symlinks is still far from foolproof. A symlink audit
;; of some sort should probably be done. E.g. if you visit the symlink
;; of a file, it will appear to be always unmodified ('cuz the symlink
;; has not in fact changed, just the contents of the file it is
;; pointing to). Viz. `vc-follow-symlinks'.
;;
;; Upon refreshing the project file (usually secondary to some other
;; action), should maybe look for affected open buffers & recalc their
;; modelines.
;;
;; Make it OK to open a p-file read-only although it needs escaping.
;; How, I don't know. Maybe turn off read-only flag and mark it as
;; unmodified?
;;
;; List major features in a user-friendly fashion.
;;
;; Caching seems to fail the first time after a parse which escaped
;; dots, though that should not make any difference since I am setting
;; the ticks after doing the parse!
;;
;; Honor `vc-make-backup-files'.
;;
;; Honor `vc-follow-symlinks' w.r.t. visit hook.
;;
;; `prcs-ask-about-save' <- `vc-suppress-confirm'. [But think about
;; effect on refreshes, which should be controllable separately.]
;;
;; `prcs-program-name' <- `vc-path'.
;;
;; Use `defined-derived-mode' if possible.
;;
;; Refresh all after a `C-x v p'.
;;
;; Search for `XXX' comments.
;;
;; ChangeLog name could be configurable. Or, could look for a file in
;; (Files) with :tag=changelog.
;;
;; ChangeLog message appearance leaves some things to be
;; desired. E.g., after inserting a message with embedded double
;; newlines, subsequent messages will be inserted in between
;; paragraphs because add-log gets confused about where it should
;; go. It may be necessary to either massage the message a fair bit
;; before insertion, or provide time for the user to do this. [See
;; discussion below.]
;;
;; Checkin should potentially see if it should merge against dest
;; branch first (if dest branch differs from current, or if acc. to
;; `prcs info -r.@ -f -q --plain-format' latest on current is not
;; this); if so, it should be able to check in onto scratch branch
;; instead (suggested: `user-login-name' + "_tmp"), then run
;; appropriate merge, then resume original checkin (unless things have
;; changed again since then...).
;;
;; What happens w/ nested projects? Presumably the subproject p-file
;; usually acts as if it belonged to the subproject, though refreshes
;; & the like may still work (modeline likely to become confused).

;; POSSIBLE VC-MODE PSEUDO-COMPATIBILITY SCHEME:
;;
;; Redefines of VC keys
;; --------------------
;; C-x C-q: default (read-only toggle) behavior
;; C-x v =: as currently in p-file, else single-file diff
;; C-x v a: see below
;; C-x v c: probably unimpl., or prompt for version deletion (p-file)
;; C-x v d: jump to p-file (currently C-x v p)
;; C-x v g: unimpl.
;; C-x v h: insert one of standard headers, or those mentioned in
;;          Project-Keywords (error from p-file), using completion; Format
;;          treated specially; just insert, no comment stuff
;; C-x v i: from p-file, create & populate new file; from other file,
;;          prompt to delete & depopulate
;; C-x v l: run info (details unclear)
;; C-x v r: checkout (file or project)
;; C-x v s: checkin (file or project)
;; C-x v u: revert to repository version (file or project); warn if mod
;; C-x v v: register (if called globally); else do a checkin
;; C-x v ~: ditto (this file, or whole project if in p-file)
;;
;; New keys in the same map
;; ------------------------
;; C-x v k: rekey (this file, or whole project)
;; C-x v n: rename (error if in p-file)
;; C-x v m: merge (just use interactive buffer, optionally with EMerge/EDiff hooks)
;; C-x v x: either release project/files (prompt if mod), or kill
;;          uncontrolled working files ("tidy", only from p-file), or
;;          both (acc. to prompt)
;; C-x v f: jump to file mentioned at point (only from p-file, else like C-x v d)
;;
;; Globally bound keys
;; -------------------
;; C-x v i: if in file buffer, and potent. PRCS controlled but not,
;;          prompt to do that; else run standard VC action
;; C-x v v: as for C-x v i (unless already in `vc-mode')
;; C-x v d: as for C-x v i (works in any buffers if potent. PRCS controlled)
;;
;; Log entry buffer
;; ----------------
;; C-c C-c: continue with check-in
;;
;; XXX should there be a way to perform an operation on more than one
;; file, but less than the whole project---or is this just overkill?

;; POSSIBLE CHANGELOG UPGRADE:
;;
;; 1. Checkin inserts New-Version-Log into ChangeLog (if any)
;; immediately after date + author + mail + filename + splat as is
;; currently done, except:
;;
;; 1a. The filename is changed to `PROJECT:VERSION' (or somesuch),
;; rather than `PROJECT.prj'. Or, the version might be put in the
;; "defun" slot.
;;
;; 1b. \ and " are unescaped.
;;
;; 1c. The body is indented over by a TAB (except blank lines) and
;; filled.
;;
;; 1d. Examine code for `vc-comment-to-change-log' for hints.
;;
;; 1e. The insertion is skipped completely if the New-Version-Log
;; begins with a pound sign.
;;
;; 1f. A new entry is always created (i.e. a new splat-header, not
;; necessarily a new date + name + address header).
;;
;; 2. An alias (indirect) buffer is created upon aborted checkin or
;; upon `C-x v a' (see later) which spans the New-Version-Log string
;; (inside the quotes). Fill mode is maybe set.
;;
;; 2a. In the alias buffer, \ and " are bound to special commands
;; which insert a previous backslash, but make that backslash
;; invisible. (The invisibility should be a text property, not an
;; overlay; disable it in the p-file with `buffer-invisibility-spec'.)
;;
;; 2b. The alias buffer is created lazily, and destroyed upon
;; reversion of the project file or checkin.
;;
;; 2c. Wherever possible, key commands from within the alias buffer
;; work as they would if called from within the p-file.
;;
;; 3. `C-x v a' is bound to a command to add a partial log entry. It
;; always jumps into the alias buffer to do its work (other-window
;; display).
;;
;; 3a. If called from p-file, it just goes to the top of the buffer:
;; i.e. the message which will appear associated with the
;; project:version itself.
;;
;; 3b. If called from a controlled file, it jumps to (the beginning
;; of) that file's entry, creating it as necessary, of the form of a
;; new paragraph, splat, filename. There may be multiple filenames
;; there already, which is fine. If called from within a defun, that
;; is added in the usual place.
;;
;; 4. Aborted checkin sets a callback in the alias buffer to continue
;; same checkin upon `C-c C-c', to mimic VC-Mode behavior.
;;
;; 5. Possibly option to only enter changelog entries for checkins
;; made on a certain set of branches (e.g. numeric branches).

(require 'lisp-mode)
(require 'emerge)
(require 'cl)
(require 'add-log)

(defvar prcs-xemacs-p (search "XEmacs" emacs-version)
  "*True if running XEmacs (vs. FSF Emacs). 20.4 currently required.")

(defvar prcs-sloppy-version-prompts nil
  "*If true, then whenever a PRCS version number (or branch) is
prompted for that is normally supposed to be an existing version
(branch), permit the user to enter any value without confirming its
existence, to save time on projects with many versions or a slow
repository. The calculation is done anyway if the user tries to use
TAB completion; this only affects typing in a value without
completion.")

(defvar prcs-check-if-file-modified t
  "*If true, check visited PRCS files to see if they have been
modified since last checkout. The check is done in the background to
avoid overhead.")

(defvar prcs-timeout-on-modified-checks 15
  "*If set to a number, that is treated as a number of seconds to wait
before killing off the PRCS status checker on a buffer, under the
assumption that it either (a) was hung, or (b) what is more likely,
Emacs totally forgot to run its sentinel & just left it.")

(defvar prcs-use-toolbar t
  "*Whether to do things to the toolbar for PRCS support. Ignored on
XEmacs.")

(defvar prcs-auto-add-changelog t
  "*If non nil and the project contains a file matching named
ChangeLog in its root directory, prcs-mode will automatically insert
the contents of New-Version-Log into the ChangeLog.")

(defvar prcs-ask-about-save t
  "*If not nil, PRCS checkin asks which buffers to save before checking in.
Otherwise, it saves all modified buffers without asking.")

(defvar prcs-program-name "prcs"
  "*Name of the prcs executable.")

(defvar prcs-repository-alist nil
  "*List of (REGEXP . REPOSITORY) pairs associating project names to
repositories. When (if any) REGEXP matches a file's project name,
REPOSITORY is used instead of the PRCS_REPOSITORY environment
variable. The first match is retained. E.g.:

'(
  (\"sharedproject\" . \"/user/sharedaccount/PRCS\")
  (\"personalproject\" . \"/user/username/PRCS\"))")

(defvar prcs-my-repository nil
  "Preferred repository to use for this project file (buffer-local).
This takes precedence over `prcs-repository-alist', so you can e.g.
set this as a buffer-local variable in the magic line of a project
file, e.g.:

-*- PRCS; prcs-my-repository: \"/foo/PRCS\" -*-")
(make-variable-buffer-local 'prcs-my-repository)

(defvar prcs-extra-checkin-options '("--long-format")
  "*Extra options passed to checkin (list of strings).")

(defvar prcs-extra-diff-args '("--context")
  "*Extra arguments passed to diff (list of strings).")

(defvar prcs-display-path t
  "*If true, display the project path to a controlled file in its modeline.")

(defvar prcs-inhibit-backups t
  "*If true, do not make backups of PRCS-controlled files (it being
redundant).")

(defvar prcs-output-buffer-divider
  ;; Don't want to actually see this in prcs.el diffs!
  (concat "------------------%<------------------%<"
	  "------------------%<------------------")
  "Divider between sections of PRCS output buffer.")

(defvar prcs-obarray-size 16383
  "Suggested size for `prcs-obarray'. Good numbers are either prime,
or 2^n-1. You only need to change this if you have a huge project, in
which case you want something on the order of 2-3 times the number of
files.")
;; ----------------------------------------------------------

;; PRCS Minor Mode.
(defvar prcs-controlled-mode nil
  "True when in a PRCS-controlled buffer.")
(make-variable-buffer-local 'prcs-controlled-mode)
;; Currently you cannot rely on these being set for some random
;; purpose; use one of the convenience functions to look up a project
;; parse &c. instead. They might not be set depending on how modeline
;; stuff is configured.
(defvar prcs-controlled-project)
(make-variable-buffer-local 'prcs-controlled-project)
(defvar prcs-controlled-version)
(make-variable-buffer-local 'prcs-controlled-version)
(defvar prcs-controlled-status)
(make-variable-buffer-local 'prcs-controlled-status)
(defvar prcs-controlled-path)
(make-variable-buffer-local 'prcs-controlled-path)

;; Project file descriptors.
(defconst prcs-descriptors
  '(Files Created-By-Prcs-Version Project-Description Project-Version Parent-Version
	  Version-Log New-Version-Log Checkin-Time Checkin-Login Populate-Ignore
	  Project-Keywords Merge-Parents New-Merge-Parents))
(defconst prcs-useful-symbols
  (mapcar 'symbol-name (append prcs-descriptors '(:symlink :no-keywords :directory)))
  "Symbols it is handy to refer to from the project file. The real
symbols are all interned privately, so simple quotes will not work.")

;; Private obarray handling.
(defvar prcs-obarray (make-vector prcs-obarray-size 0)
  "Obarray for symbols PRCS Mode creates during a parse.")
;(eval-when (compile)
;  (dolist (var prcs-useful-symbols)
;    (set (intern (concat "prcs-" var)) nil)))
(eval-when (load eval)
  (dolist (var prcs-useful-symbols)
    (set (intern (concat "prcs-" var)) (intern var prcs-obarray))))

;; Cache state.
(defvar prcs-parse-cache nil "Cached parse of a PRCS project file.")
(make-variable-buffer-local 'prcs-parse-cache)
(defvar prcs-parse-cache-tick nil "Last time a PRCS project file was parsed.")
(make-variable-buffer-local 'prcs-parse-cache-tick)

;; ---------------------------------------------------------

(defun prcs-guess-repository (pbuffer)
  "Try to guess repository from `prcs-repository-alist' or
`prcs-my-repository'."
  (save-excursion
    (set-buffer pbuffer)
    (or prcs-my-repository
	(let ((project-name (prcs-coerce (cadr (find prcs-Project-Version
						     (prcs-parse-prj-file-cached)
						     :key 'car
						     :test 'eq)))))
	  (cdr (find-if (lambda (rx)
			  (string-match rx project-name))
			prcs-repository-alist :key 'car))))))

;; ---------------------------------------------------------

;; OK to redefine, need not make use of PROJECT:
(defun prcs-output-buffer-name (project)
  "Name of the prcs output buffer, possibly based on PROJECT."
  (concat "*prcs-output [" project "]*"))

(defconst prcs-prj-descriptor-regex (mapconcat 'symbol-name prcs-descriptors "\\|")
  "Descriptors.")

(defconst prcs-font-lock-keywords
  (list
   (list (concat "(\\(" prcs-prj-descriptor-regex "\\)")
	 1 'font-lock-function-name-face)
   '("\\<:\\sw+\\>" 0 font-lock-reference-face prepend)
   )
  "PRCS keywords...")

(defconst prcs-font-lock-defaults
  `(prcs-font-lock-keywords
    nil
    t
    ,(if prcs-xemacs-p
         '((?+ . "w") (?- . "w") (?* . "w") (?/ . "w") (?. . "w") (?< . "w")
           (?> . "w") (?= . "w") (?! . "w") (?? . "w") (?$ . "w") (?% . "w")
           (?_ . "w") (?& . "w") (?~ . "w") (?^ . "w") (?: . "w"))
       '(("+-*/.<>=!?$%_&~^:" . "w")))
    beginning-of-defun
    (font-lock-comment-start-regexp . ";")
    (font-lock-mark-block-function . mark-defun))
  "PRCS font-lock defaults...")

(if prcs-xemacs-p
    ;; XEmacs
    ;; XXX broken for me---doesn't actually do anything
    (put 'prcs-mode 'font-lock-defaults prcs-font-lock-defaults)
  ;; FSF Emacs
  (eval-after-load "font-lock"
    '(or (assq 'prcs-mode font-lock-defaults-alist)
         (setq font-lock-defaults-alist
               (cons
                (cons 'prcs-mode prcs-font-lock-defaults)
                (if (boundp 'font-lock-defaults-alist)
                    font-lock-defaults-alist nil)))))
  )

(defun* prcs-check-that-file-is-ok (&key
				    (buffer (current-buffer))
				    force
				    skip-modeline)
  "Handles background disk writes by PRCS and makes sure Emacs keeps
everything in synch for BUFFER.

If the file has been written on disk since last visit/save, but is
marked unmodified here, it refreshes the buffer and sets the modeline,
after confirming with the user (unless FORCE is non-nil).

If the file has been written on disk but is modified, does the same
thing, but only after a sterner warning (which is never turned off).

If the file is untouched on disk (whether the buffer is modified or
not), this does not try to revert.

In any case, the PRCS modeline setting will be recalculated, unless a
modified buffer was left as is, or SKIP-MODELINE is true."
  (save-excursion
    (set-buffer buffer)
    (when (and (not (verify-visited-file-modtime buffer))
	       (if (buffer-modified-p)
		   (yes-or-no-p (concat "Really revert " (buffer-file-name) " from disk, discarding modifications? "))
		 (or force (y-or-n-p (concat "Refresh " (buffer-file-name) " from disk? ")))))
      (revert-buffer t t t))
    (when (and (not (buffer-modified-p))
	       (not skip-modeline))
      (prcs-update-file-status))))

(defun* prcs-get-visited-buffers (&key
				  (pbuffer (current-buffer)))
  "Return a list of all currently visited buffers corresponding to
files in project file PBUFFER (not including the project file buffer
itself)."
  ;; XXX This loops through file list and checks for the
  ;; buffer. Possibly it should actually loop through the buffers and
  ;; check for them in the file list, if that would be significantly
  ;; faster (probably unlikely to be too important unless you have a
  ;; huge project).
  ;; XXX this should skip over buffers with prcs-controlled-mode unset.
  (let ((base-dir (file-name-directory (buffer-file-name pbuffer))))
    (remove-if-not 'identity
		   (mapcar (lambda (desc)
			     (get-file-buffer (expand-file-name (prcs-coerce (car desc))
								base-dir)))
			   (cdr (find prcs-Files (prcs-parse-prj-file-cached pbuffer)
				      :key 'car))))))

(defun* prcs-prompt-for-saves (&key
			       (pbuffer (current-buffer))
			       skip-modeline
			       (buflist (cons pbuffer (prcs-get-visited-buffers :pbuffer pbuffer))))
  "For all files mentioned in the project file PBUFFER (default
current), or BUFLIST if supplied, which are also in memory buffers and
modified, save them (asking first if `prcs-ask-about-save' is
non-nil). Includes project file. If SKIP-MODELINE is set, modeline
modification checks will be suppressed."
  (save-excursion
    (dolist (buf buflist)
      (set-buffer buf)
      (when (and (buffer-modified-p)
		 (or (not prcs-ask-about-save)
		     (y-or-n-p (concat "Save file " (buffer-file-name) "? "))))
	(if skip-modeline
	    (let ((prcs-check-if-file-modified nil))
	      (save-buffer))
	  (save-buffer))))))

(defun* prcs-prompt-for-refreshes (&key
				   (pbuffer (current-buffer))
				   skip-modeline
				   (buflist (cons pbuffer (prcs-get-visited-buffers :pbuffer pbuffer))))
  "For all files mentioned in the project file PBUFFER which are also
in memory buffers, make sure they are refreshed from disk as needed
with `prcs-check-that-file-is-ok', prompting in case of
`prcs-ask-about-save'. SKIP-MODELINE used as in
`prcs-check-that-file-is-ok'. BUFLIST as in
`prcs-prompt-for-saves'. Includes p-file."
  (dolist (buf buflist)
    (prcs-check-that-file-is-ok :buffer buf :force (not prcs-ask-about-save) :skip-modeline skip-modeline)))

(defun* prcs-parse-prj-file-cached (&optional (buffer (current-buffer)))
  "Retrieve the parse for this buffer, from cache or fresh parse.
Use instead of `prcs-parse-prj-file'."
  (prcs-check-that-file-is-ok :buffer buffer :skip-modeline t)
  (save-excursion
    (set-buffer buffer)
    (when (not (boundp 'prcs-parse-cache-tick))
      (when (not (string-match "\\.prj$" (buffer-file-name)))
	(error "Attempt to parse a non-p-file %S" buffer))
      (message "Warning--forcibly converting p-file to PRCS Mode")
      (prcs-mode))
    (when (> (buffer-modified-tick) prcs-parse-cache-tick)
      (setq prcs-parse-cache (prcs-parse-prj-file buffer))
      (setq prcs-parse-cache-tick (buffer-modified-tick)))
    prcs-parse-cache))

(defun prcs-coerce (object)
  "Coerce this number, string, or symbol into a string."
  (if object
      (typecase object
	(number (number-to-string object))
	(string object)
	(symbol (symbol-name object))
	(t (error "Weird object")))))

(defun* prcs-parse-prj-file (&optional (buffer (current-buffer)))
  "Return a simple list of all s-exps in project file, as symbols,
strings, numbers etc. (according to the superficial form). No
buffer-position info retained. Lists are natural.

Note that e.g. filenames are returned as symbols, so you probably want
to use `prcs-coerce' on them; they are interned in `prcs-obarray', so
you could find one by searching (via `eq') on `(intern \"...\"
prcs-obarray)'.

To compare against a standard keyword, e.g. `New-Version-Log', use the
variable `prcs-New-Version-Log' which will contain the properly
interned symbol.

`prcs-fixup-parse' is used to perform some postprocessing.

For speed, use `prcs-parse-prj-file-cached' instead.

This format is convenient for finding things: e.g. to get checkin time as
a string, try:

  (cadr (find prcs-Checkin-Time (prcs-parse-prj-file-cached BUFFER) :key 'car))"
  (message "Parsing project file...")
  (save-excursion
    (set-buffer buffer)
    (goto-char (point-min))
    (let (prj-sexps)
      (condition-case nil
	  ;; Optimistic first stab.
	  (while t
	    (push (let ((obarray prcs-obarray)) (read buffer)) prj-sexps))
	(end-of-file nil)
	(invalid-read-syntax
	 (prcs-escape-syntactic-nastiness)
	 ;; Try again.
	 (goto-char (point-min))
	 (setq prj-sexps nil)
	 (condition-case nil
	     ;; This had better work, else an error is really thrown.
	     (while t
	       (push (let ((obarray prcs-obarray)) (read buffer)) prj-sexps))
	   (end-of-file nil))))
      (let ((result (prcs-fixup-parse
                     (nreverse prj-sexps))))
        (message "Parsing project file...done")
        result))))

(defun prcs-escape-syntactic-nastiness ()
  "Escape symbols in current project buffers beginning with
dot. Operates heuristically, so is not foolproof, but mistakes
probably won't cause any harm. Skips over anything which font-lock
claims is a comment or string."
  (let ((orig-mod (buffer-modified-p)))
    (message "Escaping filenames...")
    (goto-char (point-min))
    (while (search-forward-regexp "(\\(\\.+\\)" nil t)
      (case (get-text-property (match-beginning 1) 'face)
	(font-lock-comment-face)
	(font-lock-string-face)
	(t (goto-char (match-beginning 1))
	   (dotimes (ignore (- (match-end 1) (match-beginning 1)))
	     (insert ?\\)
	     (forward-char 1))
	   (goto-char (match-end 0)))))
    (message "Escaping filenames...done")
    (when (and (not orig-mod) (buffer-modified-p)
	       (or (not prcs-ask-about-save)
		   (y-or-n-p (concat "Save file " (buffer-file-name) "? "))))
      (save-buffer))))

(defvar prcs-careful-parses nil
  "If true, actually insert copies of new-style scoped modifiers into
all the file descriptors they apply to. Else skip this step, but still
remove them from the Files list so as not to interfere.

You only need to turn this on if you are changing code so as to
actually depend on attributes of files in a specific fashion. Leaving
it off is more efficient.")

(defun prcs-fixup-parse (parse)
  "Given a basic parse of a project file, transform it to give
the illusion that it does not have multiple file lists. Thus the parse
of the following:

    (Files :tag=haha :tag=hoho
      (foo (...))
      (bar (...) :no-keywords))
    (Files
      (baz (quux) :symlink))

will come out as if it were:

    (Files
      (foo (...) :tag=haha :tag=hoho)
      (bar (...) :no-keywords :tag=haha :tag=hoho)
      (baz (quux) :symlink))

With `prcs-careful-parses' off, this would be:

    (Files
      (foo (...))
      (bar (...) :no-keywords)
      (baz (quux) :symlink))"
  (let (files-sexps other-sexps)
    ;; Pull out Files lists from all others.
    (dolist (sexp (reverse parse))
      (if (eq (car sexp) prcs-Files)
	  (push sexp files-sexps)
	(push sexp other-sexps)))
    ;; Unified Files list will go in front, arbitrarily. Otherwise
    ;; there is no rearrangement of sexps.
    (cons (cons prcs-Files
		;; Join together all lists after processing...
		(mapcan (lambda (files-sexp)
			  (let (modifiers
				(actuals (cdr files-sexp))) ; Skip prcs-Files
			    ;; Find modifier keys.
			    (while (and actuals
					(atom (car actuals)))
			      (push (pop actuals) modifiers))
			    (setq modifiers (nreverse modifiers))
			    ;; Put them on the end of each descriptor.
			    (if (and prcs-careful-parses modifiers)
				(mapcar (lambda (desc)
					  (append desc modifiers))
					actuals)
			      ;; Speed hack:
			      actuals)))
			files-sexps))
	  other-sexps)))

;; PRCS-Emerge
;;
;; These functions allow the script 'emerge', distributed with PRCS, to
;; invoke emerge during merge.  It uses a recursive edit.  Probably should
;; know what you're doing.

(defun prcs-emerge-files (file-A file-B file-out quit-hooks)
  "Run Emerge on two files."
  (emerge-files-internal file-A file-B nil quit-hooks file-out))

(defun prcs::emerge-files-with-ancestor (file-A file-B file-ancestor file-out quit-hooks)
  "Run Emerge on two files, giving another file as the ancestor."
  (emerge-files-with-ancestor-internal file-A file-B file-ancestor nil quit-hooks file-out))

(defun prcs::emerge (work com sel out)
  "Run Emerge on two or three files."
  (message "prcs::emerge with files %s %s %s %s" work com sel out)
  (let (quit-hooks)
    (add-hook 'quit-hooks (function (lambda () (exit-recursive-edit))))
    (add-hook 'quit-hooks (` (lambda () (emerge-files-exit (, out)))))
    (if (equal com "/dev/null")
	(prcs-emerge-files work sel out quit-hooks)
      (prcs::emerge-files-with-ancestor work sel com out quit-hooks))
    (recursive-edit)
    (message "prcs::emerge finished")
    )
  )

;; This is based on lisp-mode.el

;;; Code:

(defvar prcs-mode-syntax-table lisp-mode-syntax-table)
(defvar prcs-mode-abbrev-table nil)

(define-abbrev-table 'prcs-mode-abbrev-table ())

(defun prcs-mode-variables ()
  (setq local-abbrev-table prcs-mode-abbrev-table)
  (make-local-variable 'paragraph-start)
  (setq paragraph-start (concat page-delimiter "\\|$" ))
  (make-local-variable 'paragraph-separate)
  (setq paragraph-separate paragraph-start)
  (make-local-variable 'paragraph-ignore-fill-prefix)
  (setq paragraph-ignore-fill-prefix t)
  (make-local-variable 'fill-paragraph-function)
  (setq fill-paragraph-function 'lisp-fill-paragraph)
  ;; Adaptive fill mode gets in the way of auto-fill,
  ;; and should make no difference for explicit fill
  ;; because lisp-fill-paragraph should do the job.
  (make-local-variable 'adaptive-fill-mode)
  (setq adaptive-fill-mode nil)
  (make-local-variable 'indent-line-function)
  (setq indent-line-function 'lisp-indent-line)
  (make-local-variable 'indent-region-function)
  (setq indent-region-function 'lisp-indent-region)
  (make-local-variable 'parse-sexp-ignore-comments)
  (setq parse-sexp-ignore-comments t)
  (make-local-variable 'outline-regexp)
  (setq outline-regexp ";;; \\|(....")
  (make-local-variable 'comment-start)
  (setq comment-start ";")
  (make-local-variable 'comment-start-skip)
  ;; Look within the line for a ; following an even number of backslashes
  ;; after either a non-backslash or the line beginning.
  (make-local-variable 'max-lisp-eval-depth)
  (setq max-lisp-eval-depth 10000)
  (make-local-variable 'max-specpdl-size)
  (setq max-specpdl-size 10000)
  (setq comment-start-skip "\\(\\(^\\|[^\\\\\n]\\)\\(\\\\\\\\\\)*\\);+ *")
  (make-local-variable 'comment-column)
  (setq comment-column 40)
  (make-local-variable 'comment-indent-function)
  (setq comment-indent-function 'lisp-comment-indent)
  (setq prcs-parse-cache nil)		; don't parse until requested
  (setq prcs-parse-cache-tick 0))	; ticks start at 1

(defvar prcs-mode-map nil "Keymap for `prcs-mode'.")

(defconst prcs-xemacs-menu
  '("PRCS"
    [ "Rekey"          prcs-rekey t ]
    [ "Arbitrary diff" prcs-diff-with-prefix t ]
    [ "Working diff"   prcs-diff t ]
    [ "Checkin"        prcs-checkin t ])
  "PRCS menu for XEmacs (ignored on FSF).")

(when (not prcs-mode-map)
  (if prcs-xemacs-p
      ;; XEmacs
      (progn
	;; XXX broken---adds only to buffer in effect at execution
	;; time! Needs to be done for all PRCS major-mode buffers.
        (add-submenu nil prcs-xemacs-menu)
        (setq prcs-mode-map (make-sparse-keymap))
        (set-keymap-name prcs-mode-map 'prcs-mode-map)
        (set-keymap-parents prcs-mode-map (list shared-lisp-mode-map))
        (define-key prcs-mode-map "\C-c\C-c" 'prcs-checkin)
        (define-key prcs-mode-map "\C-c\C-d" 'prcs-diff)
        (define-key prcs-mode-map "\C-c\C-r" 'prcs-rekey))
    ;; FSF Emacs
    (let ((map (make-sparse-keymap)))
      (define-key map "\C-c\C-c" 'prcs-checkin)
      (define-key map "\C-c\C-d" 'prcs-diff)
      (define-key map "\C-c\C-r" 'prcs-rekey)
      (when prcs-use-toolbar
        (define-key map [menu-bar prcs]
          (cons "PRCS" (make-sparse-keymap "PRCS")))
        (define-key map [menu-bar prcs rekey]
          '("Rekey" . prcs-rekey))
        (define-key map [menu-bar prcs complex-diff]
          '("Arbitrary diff" . prcs-diff-with-prefix))
        (define-key map [menu-bar prcs simple-diff]
          '("Working diff" . prcs-diff))
        (define-key map [menu-bar prcs checkin]
          '("Checkin" . prcs-checkin))
        )
    (setq prcs-mode-map (nconc map shared-lisp-mode-map)))
    ))

(defvar prcs-mode-hook '() "Hooks for `prcs-mode'.")

;;;###autoload
(defun prcs-mode ()
  "Major mode for editing PRCS project files.
Commands:
Delete converts tabs to spaces as it moves back.
Blank lines separate paragraphs.  Semicolons start comments.
\\{prcs-mode-map}
Entry to this mode calls the value of `prcs-mode-hook'
if that value is non-nil.
\\[prcs-checkin] checks in the project as it stands.
\\[prcs-diff] looks for differences from last checkin/checkout;
C-u \\[prcs-diff] also prompts for version numbers.
\\[prcs-rekey] rekeys the project.
\\[prcs-jump-to-project-file] jumps here from other buffers."
  (interactive)
  (kill-all-local-variables)
  (use-local-map prcs-mode-map)
  (set-syntax-table prcs-mode-syntax-table)
  (setq major-mode 'prcs-mode)
  (setq mode-name "PRCS")
  (prcs-mode-variables)
  (run-hooks 'prcs-mode-hook))
(pushnew '("\\.prj\\'" . prcs-mode) auto-mode-alist :test 'equal)

(put 'Files 'lisp-indent-function 0)
(put 'Created-By-Prcs-Version 'lisp-indent-function 0)
(put 'Project-Description 'lisp-indent-function 0)
(put 'Project-Version 'lisp-indent-function 0)
(put 'Parent-Version 'lisp-indent-function 0)
(put 'Version-Log 'lisp-indent-function 0)
(put 'New-Version-Log 'lisp-indent-function 0)
(put 'Checkin-Time 'lisp-indent-function 0)
(put 'Checkin-Login 'lisp-indent-function 0)
(put 'Populate-Ignore 'lisp-indent-function 0)
(put 'Project-Keywords 'lisp-indent-function 0)
(put 'Merge-Parents 'lisp-indent-function 0)
(put 'New-Merge-Parents 'lisp-indent-function 0)

(defun* prcs-output-buffer (&key
			    (pbuffer (current-buffer))
			    divide
			    switch-to)
  "Get a ready-to-use output buffer given PBUFFER.
If DIVIDE is `t', put in a divider at the end.
If it is `back', go back to the last divider.
If nil, just retrieve the buffer.
If SWITCH-TO is non-nil, display the buffer too."
  (let* ((orig-buf (current-buffer))
	 (project-name (prcs-coerce
			(cadr (find prcs-Project-Version
				    (prcs-parse-prj-file-cached pbuffer)
				    :key 'car))))
	 (output-buffer (get-buffer-create (prcs-output-buffer-name project-name))))
    (if switch-to (switch-to-buffer-other-window output-buffer)
      (set-buffer output-buffer))
    (goto-char (point-max))
    (cond
     ((eq divide t)
      (insert prcs-output-buffer-divider "\n")
      (goto-char (point-max)))
     ((eq divide 'back)
      (let ((pos (search-backward prcs-output-buffer-divider nil t)))
	(if pos (goto-char pos)))))
    (when (not switch-to) (set-buffer orig-buf))
    output-buffer))

;; ------------ PRCS Generic Command Interface ----------

(defvar prcs-command-locks nil
  "Locks held by running asynch processes, as alist from tokens (may
be any s-expression) to processes.")

(defvar prcs-command-queues nil
  "Queues in effect. This is an alist from queue identifiers (any
s-exp) to lists of queued-up commands to run, each of which is just
the argument list to `prcs-command' verbatim. If a queue is missing,
it is empty; if it has no elements, that means that one process is
running in it; any more elements are other processes waiting in
decreasing order of precedence.")

(defvar prcs-command-processes nil
  "All asynch processes running, as alist from processes to other
info. TIMER is the timer object if a timeout was specified, else NIL.
This other info is copied from `prcs-command' arguments:

    (WHAT REAL-REFRESH DISPLAY ERROR-ON-FAIL EXCLUSIVE SENTINEL SENTINEL-DATA SKIP-MODELINE TIMER QUEUE BUFFER PBUFFER)")

(defun prcs-command-reset ()
  "Reset PRCS command state information in case of trouble. Should not
normally be needed."
  (interactive)
  (setq prcs-command-locks nil)
  (setq prcs-command-queues nil)
  (setq prcs-command-processes nil))

;; XXX determine effect of locks & queues on mod. checks for
;; save/refresh actions implicit in command
;;
;; XXX use unwind-protect or the like to remove locks &c. as needed
;;
;; XXX handle case where command completes so quickly (thank you
;; Josh!) that Emacs is actually fooled into thinking that the file on
;; disk is not any newer than the buffer, & does not try to refresh!
;; This happened to me during a checkin. Possibly sleep for a second
;; if command was too short? But that seems a little a**-backwards.
;;
;; XXX Permit sentinels to be specified on synch commands for ease of
;; switching.
(defun* prcs-command (subcommand &rest all &key
				 options
				 (buffer (current-buffer))
				 what
				 extra-args
				 (save 'as-needed)
				 (refresh 'as-needed)
				 (display t)
				 (error-on-fail t)
				 asynch
				 (killable t)
				 exclusive
				 queue immediate
				 sentinel
				 sentinel-data
				 timeout
				 output-buffer
				 skip-modeline)
  "Run a PRCS command SUBCOMMAND with options OPTIONS (a list, `-f' is
implicit). BUFFER is a PRCS-controlled buffer which serves as the
reference point for everything. The command is always run from the
correct directory with the project file as first argument; WHAT
controls the second argument: none if `nil' (i.e. whole project),
project file if `t', or if `self', the correct relative path to BUFFER
\(even if that is the project file); EXTRA-ARGS if any are also
appended (they are not options!).

SAVE determines whether to prompt to save buffers: `nil' means none,
`t' means all visited buffers for this project, `p-file' means project
file only, list of buffers does just those (include the project file
buffer if you need it), `as-needed' (default) means acc. to value of
WHAT. Similar for REFRESH, after command termination. SKIP-MODELINE
`t' skips modeline update at end (refreshed buffers only); if
`saves-too', does not even refresh if it saves buffers. DISPLAY
\(default on) means show the output buffer (this is always written to
in any case); normally the regular PRCS output buffer is used, but
this may be overridden with OUTPUT-BUFFER. ERROR-ON-FAIL signals an
error if PRCS completed in a failure condition (status 1 does not
signal an error).

If ASYNCH is true, the command is run asynchronously (and the process
object is returned instead); KILLABLE (default true) means it can be
killed off at Emacs exit; EXCLUSIVE if set to a symbol or some
s-expression puts a kind of lock named by the EXCLUSIVE argument
\(typically containing a buffer & some symbol describing the type of
lock) and will only run if no such lock already exists (removed on
exit); SENTINEL may be given as a function accepting the process
object, the BUFFER, an exit flag (NIL for timeout, see below, or a
number for an exit status, or T for some sort of erroneous event), and
if desired SENTINEL-DATA for state, to be run upon exit. If TIMEOUT is
set, process will be stopped after that many seconds if still running,
with the sentinel being called with the timeout flag set. If QUEUE is
set to some s-expression identifying a queue, only one process may be
running in that queue at once, and the others will wait to start.

XXX Note that locks & queues currently apply only to asynchronous
commands; there is no way to lock out a synchronous command based on
them (it would only make sense in any case for locking out one
synchronous command from a set of asynchronous ones).

Returns the exit status for a synch process; the process for a
successful asynch process; for a locked-out asynch process, a list of
the existing process; or for an asynch process waiting for a queue,
nil."
  (when (and display output-buffer)
    (error "It is unsupported to DISPLAY a non-default OUTPUT-BUFFER"))
  ;; Handle queues right up front. Basically: if this process needs to
  ;; wait, put it at the end of the appropriate queue & return nil at
  ;; once. If it does not, make an empty queue so that subsequent
  ;; invocations will block correctly. Note that the internal-only
  ;; :immediate option is used to run this process anyway, since it
  ;; was just now pulled off the queue.
  (when (and asynch queue (not immediate))
    (let ((waiting (assoc queue prcs-command-queues)))
      (if waiting
	  (progn
	    ;; Note that the buffer must be specified since the
	    ;; default is context-dependent.
	    (nconc waiting (list (list* subcommand :buffer buffer all)))
	    (return-from prcs-command nil))
	(push (list queue) prcs-command-queues))))
  (let* ((filename (buffer-file-name buffer))
	 (info (prcs-is-prcs-controlled filename :get-descriptor-too t))
	 (pfilename (car info))
	 (pfilebase (file-name-nondirectory pfilename))
	 (pfiledir (file-name-directory pfilename))
	 (filenamerel (file-relative-name filename pfiledir))
	 (desc (cadr info))
	 (pbuffer (or (get-file-buffer pfilename) (find-file-noselect pfilename)))
	 (output (or output-buffer (prcs-output-buffer :pbuffer pbuffer)))
	 return-result)
  (save-excursion
    (set-buffer pbuffer)
    ;; Unless a specific list is passed in which does not contain the
    ;; p-buffer, we always want to save/refresh that!
    (flet ((interpret-s-or-r (s-or-r)
			     (if (listp s-or-r) s-or-r ; nil counts!
			       (cons pbuffer
				     (case s-or-r
				       ((p-file) nil)
				       ((t) (prcs-get-visited-buffers :pbuffer pbuffer))
				       ((as-needed)
					(case what
					  ((nil) (prcs-get-visited-buffers :pbuffer pbuffer))
					  ((t) nil)
					  ;; If buffer = p-buffer, don't repeat.
					  ((self) (if desc (list buffer)))
					  (t (error))))
				       (t (error)))))))
      (let* ((real-save (interpret-s-or-r save))
	     (real-refresh (interpret-s-or-r refresh))
	     (default-directory pfiledir)
	     (command-args (append (list subcommand "--force" "--plain-format")
				   (let ((repo (prcs-guess-repository pbuffer)))
				     (if repo (list (concat "--repository=" repo))))
				   options
				   (list pfilebase)
				   (case what
				     ((nil) nil)
				     ((t) (list pfilebase))
				     ((self) (list filenamerel))
				     (t (error "What is `what' %S?" what)))
				   extra-args))
	     (command-string (mapconcat 'identity (cons prcs-program-name command-args) " ")))
	(prcs-prompt-for-saves :pbuffer pbuffer :skip-modeline (eq skip-modeline 'saves-too) :buflist real-save)
	(unless output-buffer (prcs-output-buffer :pbuffer pbuffer :divide t))
	(if asynch
	    (progn
	      (when exclusive
		(save-excursion
		  (set-buffer buffer)
		  (let ((existing (cdr (assoc exclusive prcs-command-locks))))
		    (if existing
			(setq return-result (list existing))))))
	      (unless return-result
		(unless output-buffer
		  (set-buffer output)
		  (goto-char (point-max))
		  (insert command-string "\n")
		  (goto-char (point-max))
		  (set-buffer pbuffer))
		(let ((process (apply 'start-process
				      (append
				       (list
					command-string
					output ; XXX or (list buffer t)?
					prcs-program-name)
				       command-args))))
		  (setq return-result process)
		  (when exclusive
		    (save-excursion
		      (set-buffer buffer)
		      (push (cons exclusive process) prcs-command-locks)))
		  (set-process-sentinel process 'prcs-command-dispatch-sentinel)
		  (process-kill-without-query process (not killable))
		  (push (list process
			      what real-refresh display error-on-fail exclusive
			      sentinel sentinel-data skip-modeline
                              (if timeout
				  (if prcs-xemacs-p
				      ;; ITimer has been spectacularly unreliable for me. This works fine.
				      (add-timeout timeout 'prcs-command-dispatch-sentinel process)
				      (run-at-time timeout nil 'prcs-command-dispatch-sentinel process)))
			      queue buffer pbuffer)
			prcs-command-processes))))
	  ;; Synchronous.
	  (progn
	    (unless output-buffer
	      (when display
		(set-window-buffer (next-window nil 'no-minibuf) output))
	      (set-buffer output)
	      (goto-char (point-max))
	      (insert command-string "\n")
	      (goto-char (point-max))
	      (set-buffer pbuffer))
	    (let ((status (apply 'call-process
				 (append
				  (list prcs-program-name nil (list output (not output-buffer)) display)
				  command-args))))
	      (case status
		((0 1)
		 (setq return-result status)
		 (prcs-prompt-for-refreshes :pbuffer pbuffer :skip-modeline skip-modeline
					    :buflist real-refresh))
		(t (if error-on-fail (error "PRCS command failed: %S" status)
		     (setq return-result status)))))
	    (when (and display (not output-buffer)) (prcs-output-buffer :pbuffer pbuffer :divide 'back :switch-to t))
	    )))))
  return-result))

(defun prcs-command-dispatch-sentinel (process &optional ignore)
  "Dispatch PROCESS to the user-specified sentinel. IGNORE is the
event happening to it, which will be missing in the event of a timeout."
  (let ((timed-out (not ignore))
	(info (cdr (assq process prcs-command-processes))))
    (when (not info) (error "Process did not seem to originate from `prcs-command'!"))
    (when timed-out (message "Asynch process stopped by a timeout"))
    (let ((what (nth 0 info))		; unneeded
	  (real-refresh (nth 1 info))
	  (display (nth 2 info))
	  (error-on-fail (nth 3 info))
	  (exclusive (nth 4 info))
	  (sentinel (nth 5 info))
	  (sentinel-data (nth 6 info))
	  (skip-modeline (nth 7 info))
	  (timer (nth 8 info))
	  (queue (nth 9 info))
	  (buffer (nth 10 info))
	  (pbuffer (nth 11 info)))
      ;; Clear the timer if necessary.
      (when (and timer (not timed-out))
	(if prcs-xemacs-p
	    (disable-timeout timer)
	  (cancel-timer timer)))
      ;; Remove lock if any.
      (when exclusive
	(setq prcs-command-locks (delete* exclusive prcs-command-locks :test 'equal :key 'car)))
      ;; Remove from the process list.
      (setq prcs-command-processes (delete* process prcs-command-processes :test 'eq :key 'car))
      ;; Calculate exit status and maybe run the sentinel.
      (let ((exit-status
	     (if timed-out nil
	       (case (process-status process)
		 (exit (case (process-exit-status process)
			 (0 0)
			 (1 1)
			 (t (message "PRCS asynch process exited with an error") t)))
		 (run (message "PRCS asynch process was still running when stopped, perhaps due to timeout")
		      t)
		 (signal (message "PRCS asynch process caught some signal") t)
		 (t (message "Weird process status") t)))))
	(set-process-sentinel process nil)
	(delete-process process)
	(when (or (numberp exit-status) (not error-on-fail))
	  ;; Run the sentinel.
	  (when sentinel (apply sentinel process buffer exit-status sentinel-data))
	  ;; Perform refreshes & display.
	  (prcs-prompt-for-refreshes :pbuffer pbuffer :skip-modeline skip-modeline :buflist real-refresh)
	  (when display (prcs-output-buffer :pbuffer pbuffer :divide 'back :switch-to t)))
	;; Finally, clear the queue if applicable.
	(when queue
	  (let ((waiting (assoc queue prcs-command-queues)))
	    (when (not waiting) (error "Should have been a queue for this process!"))
	    (if (cdr waiting)
		;; There was >=1 process waiting on this one.
		(let ((next-process (cadr waiting)))
		  ;; Pull it off.
		  (setf (cdr waiting) (cddr waiting))
		  (apply 'prcs-command (append next-process '(:immediate t))))
	      ;; This queue is now free.
	      (setq prcs-command-queues (delete* queue prcs-command-queues :test 'equal :key 'car)))))
	;; Signal an error now if needed.
	(when (and error-on-fail (not (numberp exit-status)))
	  (error "PRCS asynch command failed"))))))

;; ------------------- Version Completion --------------

(defvar prcs-version-history nil
  "History list for `prcs-read-version-with-minibuffer'.")
;; Retarded language doesn't have decent closures.
(defvar *prcs-current-version-completions* nil)
(defvar *prcs-current-version-branch-only* nil)
(defvar *prcs-current-version-project* nil)
(defvar *prcs-current-version-allow-head* nil)
(defvar *prcs-current-version-default* nil)
;; XXX should have option to cache last-calculated version list (for
;; this project) on the assumption that it has not changed since then.
;; Diff could use this; currently it has to calculate it twice.
(defun* prcs-read-version-with-minibuffer
    (&key
     branch-only
     require-match
     (allow-head t)
     allow-working
     (pbuffer (prcs-find-project-file-buffer))
     (default (if allow-working ""
		(let ((version-info (find prcs-Project-Version (prcs-parse-prj-file-cached pbuffer) :test 'eq :key 'car)))
		  (if branch-only
		      (prcs-coerce (nth 2 version-info))
		    (concat (prcs-coerce (nth 2 version-info)) "." (prcs-coerce (nth 3 version-info)))))))
     (prompt (concat
	      (if require-match "Existing ")
	      "PRCS "
	      (if branch-only "branch" "version")
	      " in "
	      (prcs-coerce (nth 1 (find prcs-Project-Version (prcs-parse-prj-file-cached pbuffer) :test 'eq :key 'car)))
	      (if allow-working " (blank for working)")
	      " (TAB to complete): ")))
  "Read a PRCS version number from the minibuffer, with tab-completion
available. If BRANCH-ONLY is given, actually read a branch name, not a
version number. PROMPT is the prompt to use in the minibuffer.
REQUIRE-MATCH forces an existing PRCS version/branch to be used.
DEFAULT gives a default value; if unspecified this is taken from the
repository version of the current buffer, when in PRCS Controlled
Mode, or the working version if that is allowed. ALLOW-WORKING permits
a blank entry to be treated as the working version and accepted;
otherwise the user is not allowed to specify a blank version.
ALLOW-HEAD (default on) permits \"@\" to be used for branches or minor
versions. PBUFFER (sensibly defaulted) gives a project buffer to associate
with.

Note that REQUIRE-MATCH forces the versions to be calculated unless
ALLOW-WORKING is set and the user requests the working version, or
unless the users enters the prompted default. Otherwise this is put
off until a completion is requested by the user.

Return value is a string containing the branch/version, or `nil' for
the working version. \"@\" will be left as typed."
  (when (not pbuffer) (error "You must specify a project buffer"))
  (when (and allow-working branch-only) (error "You cannot permit a working version as a branch; this makes no sense"))
  (do ((result nil)
       (*prcs-current-version-completions* nil)
       (*prcs-current-version-project* pbuffer)
       (*prcs-current-version-branch-only* branch-only)
       (*prcs-current-version-allow-head* allow-head)
       (*prcs-current-version-default* default)
       ;; Doesn't seem to do anything:
       (minibuffer-completion-confirm t))
      ((and result (or allow-working (not (string-equal result ""))))
       (if (string-equal result "") nil result))
    (setq result (completing-read prompt 'prcs-version-completer nil
				  (and require-match (not prcs-sloppy-version-prompts))
				  default 'prcs-version-history))))
(defun prcs-version-completer (string predicate type)
  "Used by `prcs-read-version-with-minibuffer' to complete versions
\(or branches). Just does a regular completion, but does not try to
list existing versions until a completion is actually requested."
  (if (and (eq type 'lambda) (string-equal string *prcs-current-version-default*))
      t					; Quick, need not calc.
    (progn
      (unless *prcs-current-version-completions*
	(prcs-version-enumerate))
      (case type
	((nil) (try-completion string *prcs-current-version-completions* predicate))
	((t) (all-completions string *prcs-current-version-completions* predicate))
	;; XXX might be good to do a "prcs info -r<string>" to check
	;; for this case, though this is more work; doing this would
	;; probably obviate the need for
	;; `prcs-sloppy-version-prompts'. Similarly, would be possible
	;; to use "prcs info -r<string>\*" to perform
	;; completions---would this actually make anything faster?
	((lambda) (find string *prcs-current-version-completions* :test 'string-equal :key 'car))))))
(defun prcs-version-enumerate ()
  "Actually builds the list of completions for `prcs-version-completer'."
  ;; XXX ideally process filters would be used rather than a temporary
  ;; buffer, but it does not appear possible to use them on a
  ;; synchronous process, and calling wait-for-process manually etc.
  ;; seems like a pain
  (message "Getting project versions, please wait...")
  (let ((branches nil)
	(versions nil))
    (let ((output (generate-new-buffer "PRCS version info output")))
      (prcs-command "info"
		    :options '("--quiet")
		    :buffer *prcs-current-version-project*
		    :save nil
		    :refresh nil
		    :display nil
		    :skip-modeline t
		    :output-buffer output)
      (setq *prcs-current-version-completions* nil)
      (save-excursion
	(set-buffer output)
	(goto-char (point-min))
	(let ((rx (concat "^" (prcs-coerce (nth 1 (find prcs-Project-Version (prcs-parse-prj-file-cached *prcs-current-version-project*) :test 'eq :key 'car)))
			  " +\\(\\([^ ]+\\)\\.[0-9]+\\) ")))
	  (while (/= (point) (point-max))
	    (unless (looking-at rx) (error "Weird line"))
	    (push (match-string 1) versions)
	    (pushnew (match-string 2) branches :test 'string-equal)
	    (forward-line))))
      (kill-buffer output))
    (setq *prcs-current-version-completions*
	  (mapcar (lambda (x) (list x))
		  (if *prcs-current-version-branch-only* branches versions)))
    (when *prcs-current-version-allow-head*
      (if *prcs-current-version-branch-only*
	  (push '("@") *prcs-current-version-completions*)
	(dolist (branch (cons "@" branches))
	  (push (list (concat branch ".@")) *prcs-current-version-completions*)))))
  (message "Getting project versions, please wait...done"))

;; ------------- Useful PRCS Commands ------------

(defun prcs-checkin (&optional noact)
  "Checkin the current project from working dir.
Prompts for version log if necessary, and major version.
With prefix arg, takes no action, just shows what would happen."
  (interactive "P")
  (let* ((prj (prcs-parse-prj-file-cached (current-buffer)))
	 (log (cadr (find prcs-New-Version-Log prj :key 'car))))
    (when (not log)
      (goto-char (point-min))
      (error "Can't find New-Version-Log"))
    (when (and (zerop (length log))
	       (y-or-n-p "No version log entered, enter one first? "))
      (goto-char (point-min))
      ;; Best effort to find it:
      (search-forward-regexp "(New-Version-Log[ \t\n]+\"" nil t)
      (error "Checkin aborted."))
    (when (and prcs-auto-add-changelog (not noact) (not (zerop (length log))))
      (prcs-add-changelog (current-buffer) log))
    (let* ((major (prcs-coerce (caddr (find prcs-Project-Version prj :key 'car))))
	   (new-major (prcs-read-version-with-minibuffer
		       :default major
		       :allow-head t
		       :branch-only t
		       :require-match nil
		       :prompt "PRCS major version to checkin onto: "))
	   (exited (prcs-command
		    "checkin"
		    :display t
		    ;; Save time: we know what the modelines should
		    ;; be. In particular, newly added files should now
		    ;; read "", not "+", just like everything else.
		    :skip-modeline 'saves-too
		    :options (append
			      (list (concat "--revision=" new-major ".@"))
			      (if noact (list "--no-action"))
			      prcs-extra-checkin-options))))
      (let* ((new-prj (prcs-parse-prj-file-cached (current-buffer)))
	     (new-p-v (find prcs-Project-Version new-prj :key 'car))
	     (new-project (prcs-coerce (cadr new-p-v)))
	     (new-version (concat (prcs-coerce (caddr new-p-v)) "."
				  (prcs-coerce (cadddr new-p-v))))
	     (new-status ""))
	(save-excursion
	  (dolist (buf (cons (current-buffer) (prcs-get-visited-buffers)))
	    (set-buffer buf)
	    (setq prcs-controlled-project new-project)
	    (setq prcs-controlled-version new-version)
	    (setq prcs-controlled-status new-status))))
      (if (not (= exited 0))
	  (message "PRCS exited with non-zero status--failed"))
      )
    )
  )

(defun prcs-add-changelog (buffer message)
  (when (find (intern (file-name-nondirectory (change-log-name)) prcs-obarray)
	      (cdr (find prcs-Files (prcs-parse-prj-file-cached buffer) :key 'car))
	      :key 'car)
    (save-window-excursion
      (message "Adding a ChangeLog entry...")
      ;; XXX the "current defun" could plausibly be the project
      ;; version, perhaps, so you would get something like:
      ;;
      ;; foo.prj (release.17): ...
      ;;
      ;; Alternatively, this could be messily tacked onto the
      ;; name+address line above it.
      (let ((add-log-current-defun-function (lambda () nil)))
	;; XXX note that this will insert the name of the current
	;; buffer, i.e. the project file. There does not seem to be
	;; any good way to override that.
	(add-change-log-entry))
      ;; XXX should this trim outer newlines or somesuch?
      (insert message)
      ;; XXX might be nice to have it pause & wait for user to do
      ;; superficial edits, esp. as the indentation is probably
      ;; screwy.
      (save-buffer)
      (message "Adding a ChangeLog entry...done")
      )
    )
  )

(defun prcs-diff (&optional repo-versions)
  "List differences in the working version from the repository version.
With a prefix argument, prompts for versions to compare."
  (interactive "P")
  (case (prcs-command
	 "diff"
	 :extra-args (if prcs-extra-diff-args (append (list "--") prcs-extra-diff-args) nil)
	 :options (if repo-versions
		      (let* ((prj (prcs-parse-prj-file-cached (current-buffer)))
			     (pv (find prcs-Parent-Version prj :key 'car))
			     (cv (find prcs-Project-Version prj :key 'car))
			     (pv-string (prcs-read-version-with-minibuffer
					 :prompt "Older version: "
					 :default (concat (prcs-coerce (caddr pv)) "."
							  (prcs-coerce (cadddr pv)))
					 :allow-head t
					 :allow-working nil
					 :require-match t))
			     (cv-string (prcs-read-version-with-minibuffer
					 :prompt "Newer version (blank for working): "
					 :default (concat (prcs-coerce (caddr cv)) "."
							  (prcs-coerce (cadddr cv)))
					 :allow-head t
					 :allow-working t
					 :require-match t)))
			(if cv-string
			    (list
			     (concat "--revision=" pv-string)
			     (concat "--revision=" cv-string))
			  (list (concat "--revision=" pv-string))))
		    nil)
	 :skip-modeline t)
    (1 (message "There were differences."))
    (0 (message "No differences."))))
(defun prcs-diff-with-prefix ()
  "Like `prcs-diff', but with automatic prefix argument."
  (interactive)
  (prcs-diff t))

(defun prcs-rekey ()
  "Rekey the project."
  (interactive)
  (prcs-command "rekey" :skip-modeline t :display nil))

;;;;;; PRCS Minor Mode (for PRCS-controlled files, incl. the p-file)

;; Modeline display.
(or (assq 'prcs-controlled-mode minor-mode-alist)
    (push '(prcs-controlled-mode
	    (" " prcs-controlled-project ":" prcs-controlled-version
	     (prcs-controlled-status prcs-controlled-status "?")
	     prcs-controlled-path))
	  minor-mode-alist))

;; Generally already loaded into Emacs image; need `C-x v' keymap to
;; already exist:
(require 'vc-hooks)
;; Note that these bindings are global. This is actually kind of nice
;; since you can `C-x v p' from e.g. a directory listing to get to the
;; nearest applicable project file.
(global-set-key "\C-xvp" 'prcs-jump-to-project-file)

(if prcs-xemacs-p
    ;; XEmacs
    (progn
      (add-menu-button '("Tools" "VC") "----")
      (add-menu-button '("Tools" "VC")
                       [ "Jump to PRCS project" prcs-jump-to-project-file t ])
      )
  ;; FSF Emacs
  (when prcs-use-toolbar
    (define-key vc-menu-map [separator3] '("----"))
    (define-key global-map [menu-bar tools vc prcs-jump]
      '("Jump to PRCS project" . prcs-jump-to-project-file))
    ))

(add-hook 'find-file-hooks 'prcs-maybe-put-into-controlled-mode)
(add-hook 'after-save-hook 'prcs-update-file-status)

(defvar prcs-natural-b-i nil
  "Whether backups were naturally inhibited on a buffer anyway.")
(defun prcs-controlled-mode (&optional toggle info)
  "Toggle (or with prefix, turn on or off acc. to value) PRCS
Controlled Mode.

This minor mode is used for buffers that are to be considered under
PRCS control, including the project file itself (which is also in PRCS
Major Mode)."
  (interactive "P")
  (let ((old-pcm prcs-controlled-mode))
    (setq prcs-controlled-mode
	  (if toggle (> (prefix-numeric-value toggle) 0) (not old-pcm)))
    (when (and (not prcs-controlled-mode) old-pcm)
      ;; Off.
      (setq backup-inhibited prcs-natural-b-i))
    (when (and prcs-controlled-mode (not old-pcm))
      ;; We are turning it on.
      (make-local-variable 'backup-inhibited)
      (make-local-variable 'prcs-natural-b-i)
      (setq prcs-natural-b-i backup-inhibited)
      (when prcs-inhibit-backups (setq backup-inhibited t))
      (prcs-update-file-status (current-buffer) info))))

(defun* prcs-update-file-status (&optional (buffer (current-buffer)) maybe-info)
  "Update modeline status acc. to PRCS. BUFFER is buffer to check;
INFO if any is existing result of `prcs-is-prcs-controlled' with
second arg true."
  (when prcs-controlled-mode
    (save-excursion
      (set-buffer buffer)
      (let* ((info (or maybe-info (prcs-is-prcs-controlled (buffer-file-name) :get-descriptor-too t)))
	     (p-file (car info))
	     (desc (cadr info))
	     (pbuf (or (get-file-buffer p-file) (find-file-noselect p-file)))
	     (prj (prcs-parse-prj-file-cached pbuf))
	     (p-v (find prcs-Project-Version prj :key 'car)))
	(setq prcs-controlled-project (prcs-coerce (cadr p-v)))
	(setq prcs-controlled-version (concat (prcs-coerce (caddr p-v)) "."
					      (prcs-coerce (cadddr p-v))))
	(setq prcs-controlled-path
	      (if (and desc prcs-display-path)
		  (let* ((logical-path (concat "/" (prcs-coerce (car desc))))
			 (posn (search "/" logical-path :from-end t)))
		    (substring logical-path 0 (+ posn 1)))
		""))
	(setq prcs-controlled-status (if (and desc (not (cadr desc))) "+" nil))
	(when (and prcs-check-if-file-modified
		   (not prcs-controlled-status))
	  (let ((result
		 (prcs-command
		  "diff"
		  :what 'self
		  :display nil
		  :options '("--quiet")
		  :extra-args '("--" "--brief")
		  :save nil
		  :refresh nil
		  :skip-modeline t
		  :asynch t
		  :exclusive (list buffer 'modification-check)
		  :queue prcs-controlled-project
		  :timeout prcs-timeout-on-modified-checks
		  :sentinel-data nil
		  :sentinel 'prcs-update-file-status-sentinel)))
	    (cond
	     ((null result))		; waiting for the check
	     ((processp result))	; doing it now
	     ((listp result) (message "Check on %S locked out" buffer))
	     (t (message "What result is %S?!" result)))))))))
(defun prcs-update-file-status-sentinel (process buffer exit)
  (if (buffer-file-name buffer)
      (save-excursion
	(set-buffer buffer)
	(when prcs-controlled-mode
	  (setq prcs-controlled-status (case exit
					 (0 "")
					 (1 "*")
					 (t "???")))))
    (message "PRCS-controlled buffer no longer exists")))

(defun prcs-is-potentially-prcs-controlled (name)
  "Provides list of potential names of project files that could be
controlling this file. Returns nil if none, of course, so can be used
as a predicate. More specific (lower-down) project files are listed
first. (No particular order if at same level.)

XXX: could become confused if you have some controlled files in a
directory symlinked to from the actual working dir, but not present
in it. This is pretty obscure.

XXX should perhaps use `file-truename'. Full implications of symlinks
need to be worked out."
  (let* ((dir (expand-file-name (file-name-directory name)))
	 (potential nil))
    ;; Probably not the most elegant algorithm!
    (do ((idx 0)
	 (stop nil))
	(stop nil)
      (if (string-match "/" dir idx)
	  (progn
	    (setq idx (match-end 0))
	    (let* ((test-dir (substring dir 0 (1+ (match-beginning 0))))
		   (test-files (save-excursion
				 ;; ICK!!!!!!!!! Bug in directory-files: sometimes,
				 ;; depending on what buffer you are currently in,
				 ;; running directory-files dies with "Wrong type
				 ;; argument: buffer-or-string-p, t" for no apparent
				 ;; reason. Edebug claims this occurs within an
				 ;; ange-ftp hook which calls directory-files again!
				 ;; The choice of buffers in which this occurs is
				 ;; reproducible but apparently random.
				 (set-buffer (get-buffer-create "*scratch*"))
				 (directory-files test-dir t "\\.prj$"))))
	      (setq potential (nconc potential test-files))))
	(setq stop t)))
    (reverse potential)))

(defun* prcs-is-prcs-controlled (name &key get-descriptor-too)
  "Is this file listed in the (Files) section of some enclosing project
file? If so, returns the name of that project file. Lower-down project
files take precedence.

If GET-DESCRIPTOR-TOO is non-nil, will return instead a list of the project
file name, then the file descriptor list: e.g.

	(.../foo.prj (bar/baz\\.c (foo/39_bar 1.2 666) :tag=mytag))

For the project file itself, just returns that name (or a list of that
name, if GET-DESCRIPTOR-TOO is set, since there is no actual
descriptor).

XXX maybe should warn if multiple p-files contain it, to detect
possibly confusing use of subprojects."
  (if (string-match "\\.prj$" name)
      (if get-descriptor-too (list name) name)
    (save-excursion
      (block iterate
	(dolist (try (prcs-is-potentially-prcs-controlled name))
	  (let ((buf (get-file-buffer try)))
	    ;; Avoid refreshing font-lock whenever possible; it incs ticks!
	    (when (not buf) (setq buf (find-file-noselect try)))
	    ;; A little brutal, but hey...
	    (set-buffer buf)
	    (when (not (eq 'prcs-mode major-mode))
	      (prcs-mode))
	    (let ((found (find (intern
				(file-relative-name name (file-name-directory try))
				prcs-obarray)
			       (cdr (find prcs-Files
					  (prcs-parse-prj-file-cached buf)
					  :key 'car))
			       :key 'car)))
	      (when found
		(return-from iterate (if get-descriptor-too (list try found) try))))))
	nil))))

;; XXX use consistently
(defun* prcs-find-project-file-buffer (&optional (buffer (current-buffer)))
  "Find the project file (if any) controlling BUFFER."
  (let ((pfile-name (prcs-is-prcs-controlled (buffer-file-name buffer))))
    (if pfile-name (get-file-buffer pfile-name))))
(defun* prcs-find-project-file-parse (&optional (buffer (current-buffer)))
  "Find the parse of the project file (if any) controlling BUFFER."
  (let ((pbuffer (prcs-find-project-file-buffer buffer)))
    (if pbuffer (prcs-parse-prj-file-cached pbuffer))))
(defun* prcs-find-project-attribute (attribute &optional (buffer (current-buffer)))
  "Find the specified ATTRIBUTE (e.g. `prcs-Files') of the project
file controlling BUFFER. Returns whole s-exp starting with ATTRIBUTE."
  (let ((parse (prcs-find-project-file-parse buffer)))
    (if parse (find attribute parse :test 'eq :key 'car))))

(defun prcs-maybe-put-into-controlled-mode ()
  "Hook to put buffers into PRCS Controlled mode where appropriate."
  (let ((info (prcs-is-prcs-controlled (buffer-file-name) :get-descriptor-too t)))
    (when info
      (prcs-controlled-mode 1 info))))

(defun prcs-jump-to-project-file ()
  "Jump to the nearest applicable project file from here."
  (interactive)
  (let* ((here (or (buffer-file-name) default-directory))
	 (prj (or (prcs-is-prcs-controlled here)
		  (car (prcs-is-potentially-prcs-controlled here)))))
    (if prj
	(if (string-equal (buffer-file-name) prj)
	    (message "Already here!")
	  (switch-to-buffer-other-window (or (get-file-buffer prj)
					     (find-file-noselect prj))))
      (error "Could not find applicable project file!"))))

(provide 'prcs)
